#include "image.h"
#include <glm/gtc/noise.hpp>
#include "lodepng.h"
#include "tinydir.h"
#include "texture.h"

std::string getNameFromFilename(const std::string &filename)
{
    int end;
    for(end = filename.size()-1; end > 0; --end)
    {
        if(filename[end] == '.')
            break;
    }
    int start;
    for(start = end; start > 0; --start)
    {
        if(filename[start] == '/' || filename[start] == '\\')
        {
            ++start;
            break;
        }
    }
    return filename.substr(start, end-start);
}

Image::Image(const std::string &name, int myDepth, int myWidth, int myHeight, float frequency, float amplitude) :
    width(myWidth),
    height(myHeight),
    depth(myDepth),
    m_name(name)
{
    allocate(width * height * (depth/8));
    unsigned char* data = pixels.data();

    float xFactor = 1.0f / (width - 1);
    float yFactor = 1.0f / (height - 1);

    for( int row = 0; row < height; row++ ) {
        for( int col = 0 ; col < width; col++ ) {
            float x = xFactor * col;
            float y = yFactor * row;
            float sum = 0.0f;
            float freq = frequency;
            float scale = amplitude;

            // Compute the sum for each octave
            for( int oct = 0; oct < 4 ; oct++ ) {
                glm::vec2 p(x * freq, y * freq);
                float val = glm::perlin(p, glm::vec2(freq)) / scale;
                sum += val;
                float result = (sum + 1.0f)/ 2.0f;

                // Store in texture buffer
                data[((row * width + col) * (depth/8)) + oct] =
                      (unsigned char) ( result * 255.0f );
                freq *= 2.0f;   // Double the frequency
                scale *= amplitude;     // Next power of b
            }
        }
    }
}

Image::Image(const std::string &pngFile)
{
    unsigned int w, h;
    unsigned error = lodepng::decode(pixels, w, h, pngFile);
    if(error)
        fprintf(stderr, lodepng_error_text(error));
    else
    {
        width = w;
        height = h;
        depth = pixels.size()*8/(width*height);
    }
}

Image::Image(const unsigned char *pngPtr, size_t size)
{
    unsigned int w, h;
    unsigned error = lodepng::decode(pixels, w, h, pngPtr, size);
    if(error)
        fprintf(stderr, lodepng_error_text(error));
    else
    {
        width = w;
        height = h;
        depth = pixels.size()*8/(width*height);
    }
}

bool Image::save(const std::string &filename)
{
    LodePNGColorType colorType;
    switch(depth)
    {
        case 32: colorType = LCT_RGBA; break;
        case 24: colorType = LCT_RGB; break;
        case 8: colorType = LCT_GREY; break;
        default: return false; break;
    }
    unsigned error = lodepng::encode(filename, pixels, width, height, colorType, 8);
    if(error)
        fprintf(stderr, lodepng_error_text(error));
    return error == 0;
}

bool Image::save(std::vector<unsigned char> &out)
{
    LodePNGColorType colorType;
    switch(depth)
    {
        case 32: colorType = LCT_RGBA; break;
        case 24: colorType = LCT_RGB; break;
        case 8: colorType = LCT_GREY; break;
        default: return false; break;
    }
    unsigned error = lodepng::encode(out, pixels, width, height, colorType, 8);
    if(error)
        fprintf(stderr, lodepng_error_text(error));
    return error != 0;
}

void TexturePack::addImage(Image* img)
{
	TexImage texImg;
	texImg.name = img->m_name;
	texImg.img = img;
	texImg.tex = NULL;
	if(m_images.count(img->m_name))
		removeImage(img->m_name);
	m_images[img->m_name] = texImg;
}

void TexturePack::removeImage(const std::string &name)
{
	TexImage &texImg = m_images[name];
	if(texImg.img != NULL)
		delete texImg.img;
	if(texImg.tex != NULL)
		delete texImg.tex;
	m_images.erase(name);
}

void TexturePack::addImage(Texture* tex, const std::string &name, bool downloadFromGL)
{
	TexImage texImg;
	texImg.name = name;
	texImg.tex = tex;
	if(downloadFromGL)
	{
		texImg.img = tex->getData();
		texImg.img->m_name = name;
	}
	else
		texImg.img = NULL;
	if(m_images.count(name))
		removeImage(name);
	m_images[name] = texImg;
}

void TexturePack::initGL()
{
	for(std::pair<std::string, TexImage> pair : m_images)
	{
        TexImage &texImg = pair.second;
		if(texImg.img != NULL)
		{
			if(texImg.tex != NULL)
				delete texImg.tex;
			texImg.tex = new Texture(texImg.img);
		}
	}
}

void TexturePack::initDestroyImages()
{
	for(std::pair<std::string, TexImage> pair : m_images)
	{
        TexImage &texImg = pair.second;
		if(texImg.img != NULL)
		{
			if(texImg.tex != NULL)
				delete texImg.tex;
			texImg.tex = new Texture(texImg.img);
			delete texImg.img;
		}
	}
}

void TexturePack::getImagesFromGL()
{
	for(std::pair<std::string, TexImage> pair : m_images)
	{
        TexImage &texImg = pair.second;
		if(texImg.tex != NULL)
		{
			if(texImg.img != NULL)
				delete texImg.img;
			texImg.img = texImg.tex->getData();
		}
	}
}

void TexturePack::destroyGL()
{
	for(std::pair<std::string, TexImage> pair : m_images)
	{
        TexImage &texImg = pair.second;
		if(texImg.tex != NULL)
			delete texImg.tex;
	}
}

bool TexturePack::load(const std::string &folderPath)
{
	m_name = getNameFromFilename(folderPath);
	tinydir_dir dir;
	int err = tinydir_open(&dir, folderPath.c_str());
	if(err == -1)
	{
		fprintf(stderr, "can't open directory : \"%s\"\n", folderPath.c_str());
		return false;
	}
	while(dir.has_next)
	{
		tinydir_file file;
        tinydir_readfile(&dir, &file);
        if(std::string(file.name) == "." || std::string(file.name) == ".." || file.is_dir)
        {
            tinydir_next(&dir);
            continue;
        }
		if(std::string(file.extension) == "png")
		{
			Image* img = new Image(std::string(file.path));
			img->m_name = file.name;
			addImage(img);
		}
		else
			fprintf(stderr, "%s is not a png file, ignoring\n", file.path);
        tinydir_next(&dir);
	}
	return true;
}

bool TexturePack::save(const std::string &folderPath)
{
	std::string path = folderPath + '/' + m_name;
	tinydir_dir dir;
	int err = tinydir_open(&dir, path.c_str());
	if(err == -1)
	{
        if(!tinydir_create_directory(path.c_str()))
        {
            fprintf(stderr, "can't create directory \"%s\"\n", path.c_str());
            return false;
        }
	}
	err = tinydir_open(&dir, path.c_str());
	if(err == -1)
	{
		fprintf(stderr, "can't open directory : \"%s\"\n", path.c_str());
		return false;
	}
	// directory is open
	for(std::pair<std::string, TexImage> pair : m_images)
	{
        TexImage &texImg = pair.second;
        if(texImg.img != NULL)
        {
            std::string filename = path + '/' + texImg.name;
            if(!texImg.img->save(filename))
                fprintf(stderr, "can't save image : \"%s\"\n", filename.c_str());
        }
	}
}
